/*
 * Copyright © 2006 Carl Worth
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA."
 */

/* for asprintf */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include <math.h>

#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <string.h>
#include <assert.h>

#include <readline/readline.h>
#include <readline/history.h>

typedef int bool_t;

typedef struct _item {
    int score;
    char *challenge;
    char *response;
} item_t;

typedef struct _bin {
    int score;
    int items_size;
    int num_items;
    item_t **items;
} bin_t;

typedef enum {
    CATEGORY_ORDER_RANDOM,
    CATEGORY_ORDER_SEQUENTIAL
} category_order_t;

typedef struct _category {
    char *name;
    int items_size;
    int num_items;
    item_t *items;

    /* Support sequential introduction of items from bin 0 */
    category_order_t order;
    int bin_zero_head;
} category_t;

typedef struct _mnemon {
    char *dir_name;

    int categories_size;
    int num_categories;
    category_t *categories;

    int bins_size;
    int num_bins;
    bin_t *bins;

    int to_introduce;
    int to_master;
    int unlearned;
    int mastered;
} mnemon_t;

static void *
xmalloc (size_t size)
{
    void *ret;

    ret = malloc (size);
    if (ret == NULL) {
	fprintf (stderr, "Error: out of memory\n");
	exit (1);
    }

    return ret;
}

static void *
xrealloc (void *ptr, size_t size)
{
    void *ret;

    ret = realloc (ptr, size);
    if (ret == NULL) {
	fprintf (stderr, "Error: out of memory\n");
	exit (1);
    }

    return ret;
}

static char *
xstrdup (const char *s)
{
    char *ret;

    ret = strdup (s);
    if (ret == NULL) {
	fprintf (stderr, "Error: out of memory\n");
	exit (1);
    }

    return ret;
}

static char *
xstrndup (const char *s, size_t n)
{
    char *ret;

    ret = strndup (s, n);
    if (ret == NULL) {
	fprintf (stderr, "Error: out of memory\n");
	exit (1);
    }

    return ret;
}

static void
xasprintf (char **strp, const char *fmt, ...)
{
    va_list ap;
    int ret;

    va_start (ap, fmt);
    ret = vasprintf (strp, fmt, ap);
    va_end (ap);

    if (ret < 0) {
	fprintf (stderr, "Error: out of memory\n");
	exit (1);
    }
}

static void
item_init (item_t	*item,
	   int		 score,
	   const char	*challenge,
	   const char	*response)
{
    item->score = score;

    item->challenge = xmalloc (strlen (challenge) + 1 +
			       strlen (response) + 1);
    item->response = item->challenge + strlen (challenge) + 1;

    strcpy (item->challenge, challenge);
    strcpy (item->response, response);
}

static void
item_fini (item_t *item)
{
    /* item->response shares allocation with item->challenge, so
     * doesn't require a separate call to free */
    free (item->challenge);
}

static void
category_init (category_t *category,
	       const char *name)
{
    category->name = xstrdup (name);
    
    category->items_size = 0;
    category->num_items = 0;
    category->items = NULL;
    category->order = CATEGORY_ORDER_RANDOM;
    category->bin_zero_head = 0;
}

static void
category_fini (category_t *category)
{
    int i;

    for (i = 0; i < category->num_items; i++)
	item_fini (&category->items[i]);

    free (category->items);

    free (category->name);
}

static void
category_grow (category_t *category)
{
    if (category->items_size)
	category->items_size *= 2;
    else
	category->items_size = 1;

    category->items = xrealloc (category->items,
				category->items_size * sizeof (item_t));
}

static item_t *
category_add_item (category_t	*category,
		   int		 score,
		   const char	*challenge,
		   const char	*response)
{
    item_t *item;

    if (category->num_items == category->items_size)
	category_grow (category);

    item = &category->items[category->num_items++];

    item_init (item, score, challenge, response);

    return item;
}

static item_t *
category_next_bin_zero_item (category_t	*category)
{
    int *i = &category->bin_zero_head;

    for ( ; *i < category->num_items; *i = *i + 1)
	if (category->items[*i].score == 0)
	    return &category->items[*i];

    return NULL;
}

static void
category_print (category_t	*category,
		FILE		*file)
{
    int i;
    item_t *item;

    fprintf (file, "order = %s\n\n",
	    category->order == CATEGORY_ORDER_RANDOM ? "random" : "sequential");

    for (i = 0; i < category->num_items; i++) {
	item = &category->items[i];
	if (i != 0)
	    fprintf (file, "\n");
	fprintf (file, "%d\n%s\n%s\n",
		 item->score,
		 item->challenge,
		 item->response);
    }
}

static void
bin_init (bin_t	*bin,
	  int	 score)
{
    bin->score = score;

    bin->items_size = 0;
    bin->num_items = 0;
    bin->items = NULL;
}

static void
bin_fini (bin_t *bin)
{
    free (bin->items);
}

static void
bin_grow (bin_t *bin)
{
    if (bin->items_size)
	bin->items_size *= 2;
    else
	bin->items_size = 1;

    bin->items = xrealloc (bin->items,
			   bin->items_size * sizeof (item_t*));
}

static void
bin_add_item (bin_t 	*bin,
	      item_t	*item)
{
    assert (item->score == bin->score);

    if (bin->num_items == bin->items_size)
	bin_grow (bin);

    bin->items[bin->num_items++] = item;
}

static void
bin_remove_item (bin_t	*bin,
		 int	 item_index)
{
    /* Replace the current item with the last item, (no need to shift
     * any more than that since we don't care about the order of the
     * items within a bin). */
    bin->num_items--;
    if (bin->num_items)
	bin->items[item_index] = bin->items[bin->num_items];
}

/* Find the index for an item within a bin.
 *
 * XXX: This is currently a linear search, so is a potential
 * performance problem.
 */
static int
bin_item_index (bin_t	*bin,
		item_t	*item)
{
    int i;

    for (i = 0; i < bin->num_items; i++)
	if (bin->items[i] == item)
	    return i;

    assert (0);
}

typedef int (item_match_predicate_t) (void *closure, item_t *item);

/* Return the number of items in the bin from the given category (or
 * from all categories if category == NULL) */
static int
bin_num_items_matching (bin_t			*bin,
			item_match_predicate_t	*predicate,
			void			*closure)
{
    int i, num_items = 0;

    if (predicate == NULL)
	return bin->num_items;

    for (i = 0; i < bin->num_items; i++)
	if ((predicate) (closure, bin->items[i]))
	    num_items++;

    return num_items;
}

static void
mnemon_init (mnemon_t *mnemon)
{
    char *home;

    home = getenv ("HOME");
    if (home == NULL)
	home = "";

    xasprintf (&mnemon->dir_name, "%s/.mnemon", getenv ("HOME"));

    mnemon->categories_size = 0;
    mnemon->num_categories = 0;
    mnemon->categories = NULL;

    mnemon->bins_size = 0;
    mnemon->num_bins = 0;
    mnemon->bins = NULL;

    mnemon->to_introduce = 10;
    mnemon->to_master = 0;
    mnemon->unlearned = 0;
    mnemon->mastered = -1;
}

static void
mnemon_fini (mnemon_t *mnemon)
{
    int i;

    for (i = 0; i < mnemon->num_bins; i++)
	bin_fini (&mnemon->bins[i]);
    free (mnemon->bins);

    for (i = 0; i < mnemon->num_categories; i++)
	category_fini (&mnemon->categories[i]);
    free (mnemon->categories);

    free (mnemon->dir_name);
}

static void
mnemon_categories_grow (mnemon_t *mnemon)
{
    if (mnemon->categories_size)
	mnemon->categories_size *= 2;
    else
	mnemon->categories_size = 1;

    mnemon->categories = xrealloc (mnemon->categories,
				   mnemon->categories_size * sizeof (category_t));
}

/* Get a category by name if it exists */
static category_t *
mnemon_get_category_if_exists (mnemon_t	    *mnemon,
			       const char   *name)
{
    int i;

    for (i = 0; i < mnemon->num_categories; i++)
	if (strcmp (mnemon->categories[i].name, name) == 0)
	    return &mnemon->categories[i];

    return NULL;
}

/* Get a category by name, creating new one if necessary. */
static category_t *
mnemon_get_category (mnemon_t	*mnemon,
		     const char	*name)
{
    category_t *category;

    category = mnemon_get_category_if_exists (mnemon, name);
    if (category)
	return category;

    mnemon_categories_grow (mnemon);

    category = &mnemon->categories[mnemon->num_categories++];

    category_init (category, name);

    return category;
}

static void
mnemon_bins_grow (mnemon_t *mnemon)
{
    if (mnemon->bins_size)
	mnemon->bins_size *= 2;
    else
	mnemon->bins_size = 1;

    mnemon->bins = xrealloc (mnemon->bins,
			     mnemon->bins_size * sizeof (bin_t));
}

static bin_t *
mnemon_get_bin (mnemon_t	*mnemon,
		int		 score)
{
    int i;
    bin_t *bin;

    for (i = 0; i < mnemon->num_bins; i++)
	if (mnemon->bins[i].score == score)
	    return &mnemon->bins[i];
	else if (mnemon->bins[i].score > score)
	    break;

    if (mnemon->num_bins == mnemon->bins_size)
	mnemon_bins_grow (mnemon);

    bin = &mnemon->bins[i];

    /* Make room to insert new bin at its sorted location. */
    if (i < mnemon->num_bins)
	memmove (bin + 1, bin, (mnemon->num_bins - i) * sizeof (bin_t));
    mnemon->num_bins++;

    bin_init (bin, score);

    return bin;
}

static void
mnemon_remove_bin (mnemon_t	*mnemon,
		   bin_t	*bin)
{
    int i = bin - mnemon->bins;

    bin_fini (bin);

    memmove (bin, bin + 1, (mnemon->num_bins - i) * sizeof (bin_t));
    mnemon->num_bins--;
}

static void
chomp (char *s)
{
    int len = strlen (s);
    if (len == 0)
	return;
    if (s[len - 1] == '\n')
	s[len - 1] = '\0';
}

static char *
trim_space (char *string)
{
    char *s;

    s = string;
    while (*s && isspace (*s))
	s++;

    string = s;

    s = string + strlen (string) - 1;
    while (s > string && isspace (*s)) {
	*s = '\0';
	s--;
    }

    return string;
}

static void
mnemon_load_category (mnemon_t		*mnemon,
		      const char	*name)
{
    FILE *file;
    char *line = NULL, *end;
    size_t line_size = 0;
    ssize_t bytes_read;
    int line_count = 0;
    char *path;
    category_t *category;
    int i;

    path = xmalloc (strlen (mnemon->dir_name) + 1 + strlen (name) + 1);
    sprintf (path, "%s/%s", mnemon->dir_name, name);

    file = fopen (path, "r");
    if (file == NULL) {
	fprintf (stderr, "Error: Failed to open %s: %s\n",
		 path, strerror (errno));
	exit (1);
    }

    category = mnemon_get_category (mnemon, name);

#define READ_LINE do {					\
    bytes_read = getline (&line, &line_size, file);	\
    if (bytes_read == -1)				\
	goto END_OF_FILE;				\
    line_count++;					\
    chomp (line);					\
} while (0)

    /* Parse options */
    while (1) {
	char *name, *equal, *value;

	/* Ignore blank lines */
	READ_LINE;
	if (*line == '\0')
	    continue;

	/* An initial digit means we hit an item. Trigger the
	 * spaghetti machine. */
	if (*line >= '0' && *line <= '9')
	    goto PARSE_BIN;

	equal = strchr (line, '=');
	if (equal == NULL) {
	    fprintf (stderr, "Malformed option, (expected name=value): \"%s\" at %s:%d\n",
		     line, path, line_count);
	    exit (1);
	}

	value = equal + 1;
	name = line;
	*equal = '\0';

	name = trim_space (name);
	value = trim_space (value);

	if (strcmp (name, "order") == 0) {
	    if (strcmp (value, "sequential") == 0) {
		category->order = CATEGORY_ORDER_SEQUENTIAL;
	    } else if (strcmp (value, "random") == 0) {
		category->order = CATEGORY_ORDER_RANDOM;
	    } else {
		fprintf (stderr, "Unknown value for \"order\" option \"%s\" at %s:%d\n",
			 value, path, line_count);
		exit (1);
	    }
	} else {
	    fprintf (stderr, "Unknown option %s at %s:%d\n",
		     name, path, line_count);
	    exit (1);
	}
    }

    /* Parse items */
    while (1) {
	int score;
	char *challenge, *response;

	/* Ignore blank lines */
	READ_LINE;
	if (*line == '\0')
	    continue;

	/* Read bin number */
      PARSE_BIN:
	score = strtol (line, &end, 10);
	if (*end != '\0') {
	    fprintf (stderr, "Failed to parse bin number from \"%s\" at %s:%d\n",
		     line, path, line_count);
	    exit (1);
	}

	/* Read challenge */
	READ_LINE;
	challenge = strdup (line);

	/* Read response */
	READ_LINE;
	response = line;

	category_add_item (category, score, challenge, response);

	free (challenge);
    }
  END_OF_FILE:

    free (line);
    fclose (file);
    free (path);

    /* Resize category items to fit exactly. */
    category->items_size = category->num_items;
    category->items = xrealloc (category->items, category->items_size * sizeof (item_t));

    /* Now that the category is completely loaded, with stable
     * pointers to every item, we can add each item to its appropriate
     * bin. */
    for (i = 0; i < category->num_items; i++) {
	item_t *item = &category->items[i];
	bin_t *bin = mnemon_get_bin (mnemon, item->score);

	bin_add_item (bin, item);
    }
}

static void
mnemon_load (mnemon_t *mnemon)
{
    DIR *dir;
    struct dirent *dirent;

    dir = opendir (mnemon->dir_name);
    if (dir == NULL) {
	fprintf (stderr, "Error: Failed to open directory %s: %s\n",
		 mnemon->dir_name, strerror (errno));
	exit (1);
    }

    while (1) {
	dirent = readdir (dir);
	if (dirent == NULL)
	    break;

	if (dirent->d_type == DT_REG) {
	    /* Ignore files matching *~, (yes, this shouldn't be
	     * hard-coded in such an ad-hoc way, but there you go. */
	    if (dirent->d_name[strlen(dirent->d_name)-1] != '~')
		mnemon_load_category (mnemon, dirent->d_name);
	}
    }

    closedir (dir);
}

static void
mnemon_save (mnemon_t *mnemon)
{
    int i, err;
    char *filename, *lock_filename;
    FILE *file;
    category_t *category;

    for (i = 0; i < mnemon->num_categories; i++) {
	category = &mnemon->categories[i];

	xasprintf (&filename, "%s/%s",
		   mnemon->dir_name, category->name);
	xasprintf (&lock_filename, "%s/.#%s",
		   mnemon->dir_name, category->name);

	file = fopen (lock_filename, "w");
	if (file == NULL) {
	    fprintf (stderr, "Error: Failed to open %s for writing: %s\n",
		     lock_filename, strerror (errno));
	    continue;
	}

	category_print (category, file);

	fclose (file);

	err = rename (lock_filename, filename);
	if (err < 0) {
	    fprintf (stderr, "Error: Failed to rename %s to %s: %s\n",
		     lock_filename, filename, strerror (errno));
	    continue;
	}

	free (filename);
	free (lock_filename);
    }
}

/* Return a uniformly-distributed pseudo-random integer within the
 * range:
 *
 *	0 <= result < num_values
 */
static int
rand_within (int num_values)
{
    return (int) (num_values * (rand() / (RAND_MAX + 1.0)));
}

/* Return an exponentially-distributed pseudo-random integer within
 * the range:
 *
 *	0 <= result < num_values
 *
 * The distribution is such that each successively larger value will
 * occur with a probability of half of the previous value.
 */
static int
rand_within_exponential (int num_values)
{
    static int r;
    static uint32_t mask = 0;
    int ones;
    int bit;

    /* Optimize the constant case. */
    if (num_values == 1)
	return 0;

    ones = 0;

    do {
	if (mask == 0) {
	    r = rand ();
	    mask = 1 << 31;
	    while (mask > RAND_MAX)
		mask >>= 1;
	}
	bit = r & mask;
	mask >>= 1;
	if (bit) {
	    ones++;
	    if (ones == num_values)
		ones = 0;
	}
    } while (bit);

    return ones;
}

/* Find the category to which an item belongs. */
static category_t *
mnemon_item_category (mnemon_t	*mnemon,
		      item_t	*item)
{
    category_t *category;
    int i, item_index;

    for (i = 0; i < mnemon->num_categories; i++) {
	category = &mnemon->categories[i];
	item_index = item - category->items;
	if (item_index >= 0 && item_index < category->num_items)
	    return category;
    }

    assert (0);
}

typedef struct _item_in_category_closure
{
    mnemon_t *mnemon;
    category_t *category;
} item_in_category_closure_t;

static int
mnemon_item_in_category (void *closure, item_t *item)
{
    item_in_category_closure_t *iicc = closure;
    mnemon_t *mnemon = iicc->mnemon;
    category_t *category = iicc->category;

    return (mnemon_item_category (mnemon, item) == category);
}

typedef struct _item_in_category_of_length_closure
{
    mnemon_t *mnemon;
    category_t *category;
    int length;
} item_in_category_of_length_closure_t;

static int
mnemon_item_in_category_of_length (void *closure, item_t *item)
{
    item_in_category_of_length_closure_t *iicolc = closure;
    mnemon_t *mnemon = iicolc->mnemon;
    category_t *category = iicolc->category;
    int length = iicolc->length;

    if (mnemon_item_category (mnemon, item) != category)
	return 0;

    return strlen (item->challenge) == length;
}

static void
mnemon_select_item (mnemon_t	 *mnemon,
		    bin_t	**bin_ret,
		    int		 *item_index_ret)
{
    int bin_index, item_index;
    bin_t *bin;

    bin_index = rand_within_exponential (mnemon->num_bins);

    bin = &mnemon->bins[bin_index];

    item_index = rand_within (bin->num_items);

    if (bin->score == 0) {
	category_t *category;
	item_t *item;

	item = bin->items[item_index];

	category = mnemon_item_category (mnemon, item);

	if (category->order == CATEGORY_ORDER_SEQUENTIAL) {
	    item = category_next_bin_zero_item (category);
	    if (item)
		item_index = bin_item_index (bin, item);
	}
    }

    *bin_ret = bin;
    *item_index_ret = item_index;
}


#define HISTOGRAM_ROW_FORMAT "%3d: %3d"
#define HISTOGRAM_BAR_WIDTH  63

static void
print_histogram_bar (double	size,
		     double	max)
{
    int units_per_cell = (int) ceil (max / HISTOGRAM_BAR_WIDTH);
    static char const *boxes[8] = {
	"█", "▉", "▊", "▋",
	"▌", "▍", "▎", "▏"
    };

    while (size > units_per_cell) {
	printf(boxes[0]);
	size -= units_per_cell;
    }

    size /= units_per_cell;

    if (size > 7.5/8.0)
	printf(boxes[0]);
    else if (size > 6.5/8.0)
	printf(boxes[1]);
    else if (size > 5.5/8.0)
	printf(boxes[2]);
    else if (size > 4.5/8.0)
	printf(boxes[3]);
    else if (size > 3.5/8.0)
	printf(boxes[4]);
    else if (size > 2.5/8.0)
	printf(boxes[5]);
    else if (size > 1.5/8.0)
	printf(boxes[6]);
    else if (size > 0.5/8.0)
	printf(boxes[7]);

    printf ("\n");
}

static void
mnemon_print_histogram (mnemon_t    *mnemon,
			const char  *category_name,
			int	     length)
{
    int i, last_score, max;
    category_t *category = NULL;
    bin_t *bin;
    int num_items;
    item_match_predicate_t *predicate = NULL;
    void *closure = NULL;
    item_in_category_closure_t item_in_category;
    item_in_category_of_length_closure_t item_in_category_of_length;

    if (mnemon->num_bins == 0)
	return;

    if (category_name) {
	category = mnemon_get_category_if_exists (mnemon, category_name);
	if (category) {
	    if (length) {
		predicate = mnemon_item_in_category_of_length;
		item_in_category_of_length.mnemon = mnemon;
		item_in_category_of_length.category = category;
		item_in_category_of_length.length = length;
		closure = &item_in_category_of_length;
	    } else {
		predicate = mnemon_item_in_category;
		item_in_category.mnemon = mnemon;
		item_in_category.category = category;
		closure = &item_in_category;
	    }
	}
    }

    for (i = 0; i < mnemon->num_bins; i++) {
	num_items = bin_num_items_matching (&mnemon->bins[i],
					    predicate, closure);
	if (i == 0 || num_items > max)
	    max = num_items;
    }

    for (i = 0; i < mnemon->num_bins; i++) {
	bin = &mnemon->bins[i];
	if (i != 0)
	    while (bin->score - last_score > 1)
		printf (HISTOGRAM_ROW_FORMAT "\n", ++last_score, 0);
	num_items = bin_num_items_matching (bin,
					    predicate, closure);
	printf (HISTOGRAM_ROW_FORMAT " ", bin->score, num_items);
	print_histogram_bar (num_items, max);
	last_score = bin->score;
    }
}

static void
mnemon_handle_command (mnemon_t		*mnemon,
		       const char	*command)
{
    const char *arg;
    int len;
    switch (command[0]) {
	case 'h':
	{
	    char *category = NULL;
	    int length = 0;

	    arg = command + 1;
	    arg += strspn (arg, " \t");
	    len = strcspn (arg, " \t");
	    if (len) {
		category = xstrndup (arg, len);
		arg += len;
		arg += strspn (arg, " \t");
		if (*arg)
		    length = atoi (arg);
	    }
	    mnemon_print_histogram (mnemon, category, length);
	}
	break;
        default:
	    printf ("Unknown command: %s\n", command);
	    break;
    }
}

static void
mnemon_handle_response (mnemon_t	*mnemon,
			bin_t		*bin,
			int		 item_index,
			item_t		*item,
			const char	*response)
{
    bool_t correct;

    correct = (strcmp (response, item->response) == 0);

    bin_remove_item (bin, item_index);

    /* If the bin is now empty, we must remove it. Also if we just
     * picked the last word we'll ever pick from the bin with
     * score 0, then we can remove that as well. */
    if (bin->num_items == 0 ||
	(bin->score == 0 && mnemon->to_introduce == 0))
    {
	mnemon_remove_bin (mnemon, bin);
    }

    if (correct) {
	item->score++;
	/* We reserve an item score of 0 for an item that has
	 * never been asked. */
	if (item->score == 0) {
	    item->score = 1;
	    mnemon->unlearned--;
	    printf ("You got it!");
	} else if (item->score < 0) {
	    printf ("Yes---just give me %d more.",
		    - item->score);
	} else if (item->score == 1) {
	    printf ("On your first try, no less!");
	} else {
	    printf ("Masterful (%dx).", item->score);
	    if (mnemon->to_master)
		mnemon->mastered++;
	}
    } else {
	printf ("  %s is the correct answer.",
		item->response);
	/* Penalize an incorrect response by forcing the score
	 * negative. */
	if (item->score >= 0) {
	    if (item->score > 0)
		printf (" Oops, you knew that, right? (%d x)\n ",
			item->score);
	    mnemon->unlearned++;
	    /* We go to -2 to force a little extra reinforcement
	     * when re-learning an item, (otherwise, it will often
	     * get asked again immediately where it is easy to get
	     * a correct response without any learning). */
	    item->score = -2;
	} else {
	    item->score--;
	}
    }

    if (mnemon->to_introduce == 0 &&
	mnemon->unlearned == 0 &&
	mnemon->to_master == 0)
    {
	mnemon->to_master = 10;
	mnemon->mastered = 0;
    }

    printf (" ");
    if (mnemon->to_introduce)
	printf ("%d to come. ", mnemon->to_introduce);
    if (mnemon->unlearned)
	printf ("%d still unlearned. ", mnemon->unlearned);
    if (mnemon->to_master) {
	if (mnemon->mastered < mnemon->to_master)
	    printf ("%d items to master",
		    mnemon->to_master - mnemon->mastered);
	else
	    printf ("Great job!");
    }
    printf ("\n\n");

    bin = mnemon_get_bin (mnemon, item->score);

    bin_add_item (bin, item);
}

static void
mnemon_do_challenges (mnemon_t *mnemon)
{
    bin_t *bin;
    int item_index;
    item_t *item;
    char *response;
    int i;

    /* Count the number of items with negative scores. */
    mnemon->unlearned = 0;
    for (i = 0; i < mnemon->num_bins; i++) {
	bin = &mnemon->bins[i];
	if (bin->score >= 0)
	    break;
	mnemon->unlearned += bin->num_items;
    }

    mnemon->to_introduce -= mnemon->unlearned;
    if (mnemon->to_introduce < 0)
	mnemon->to_introduce = 0;

    /* Get rid of bin with score of 0 if we aren't going to be
     * introducing anything from it. */
    if (mnemon->to_introduce == 0) {
	bin = mnemon_get_bin (mnemon, 0);
	mnemon_remove_bin (mnemon, bin);	
    }

    if (mnemon->unlearned) {
	printf ("You've got %d items to learn already. ", mnemon->unlearned);
	if (mnemon->to_introduce)
	    printf ("I'll introduce %d more as we go.", mnemon->to_introduce);
	printf ("\n");
    } else {
	printf ("Introducing %d new items.\n", mnemon->to_introduce);
    }
    printf ("\n");

    do {
	mnemon_select_item (mnemon, &bin, &item_index);
	item = bin->items[item_index];

	if (bin->score == 0)
	    mnemon->to_introduce--;

	while (1) {
	    printf ("%s\n", item->challenge);

	    response = readline ("> ");
	    /* Terminate on EOF */
	    if (response == NULL) {
		printf ("\n");
		return;
	    }

	    if (response[0] == '/')
		mnemon_handle_command (mnemon, response + 1);
	    else
		break;
	}

	mnemon_handle_response (mnemon, bin, item_index,
				item, response);
    } while (mnemon->mastered < mnemon->to_master);
}

int
main (int argc, char *argv[])
{
    mnemon_t mnemon;

    srand (time (NULL));

    mnemon_init (&mnemon);

    mnemon_load (&mnemon);

    mnemon_do_challenges (&mnemon);

    mnemon_save (&mnemon);

    mnemon_fini (&mnemon);

    return 0;
}
